* MicroEMACS screen driver (c) 1992      Richard Kettlewel
*                          (c) 1993-1999 Thierry Godefroy.
*
* This contains the assembly code part of MicroEMACS 3.11 QL screen driver.
* Although it is not necessary to write it in assembly (earlier versions
* were in C) it would be difficult to make it much faster than the assembly
* version.
*
* Additionally, the assembly version is smaller than the C version.
*
*               21/03/92        Wrote assembly code.
*               28/05/92        Added _typahead.
*               29/05/92        It all works (now).
*               31/05/92        Transferred some more stuff from ql_c.
*               01/06/92        More transfers + a bug removed.
*               02/06/92        Minor debugging.
*
* Modifications and additions by Thierry Godefroy:
*
*               03/04/93        qlmouse() added (read mouse displacement).
*               04/11/99        Changed qlflush() so that no attempt is made to
*                                call the screen driver when the emacs window
*                                is locked by the pointer environment (instead
*                                the buffer is emptied and the screen is marked
*                                as garbage so that a redraw is triggered on
*                                the first occasion). As a result, a test has
*                                been added in all _ql* routines after calling
*                                _qlflush in order to check for an error and
*                                abort if there is one.
*               18/11/99        Character attributes support added (new qlbold()
*                                qitalic() and qlunder() functions added).
*

*
* Keys
*
forever  equ    -1
mt.inf   equ    $00
io.pend  equ    $00
io.sstrg equ    $07
sd.extop equ    $09
sd.cure  equ    $0e
sd.pos   equ    $10
sd.tab   equ    $11
sd.pcol  equ    $13
sd.clrbt equ    $22
sd.clrrt equ    $24
sd.fount equ    $25
sd.setpa equ    $27
sd.setst equ    $28
sd.setin equ    $29
sd.setul equ    $2b
iop.pick equ    $7c
jb_tag   equ    $10
sv_jbmax equ    $62
sv_jbbas equ    $68
sv_chbas equ    $78

*
* Static variables
*
        .bss
qlbuffer:
        ds.b    100                             ; Output buffer

        .data
pointer:
        dc.l    qlbuffer                        ; Pointer into output buffer
remains:
        dc.w    100                             ; Bytes remaining in the buffer
_insubmenu:                                     ; Flag set when the FSEL or
        dc.b    0                               ; "variables" sub-menu is used.

*
* Code
*
        .text
*
* int qlflush(void)
* First checks for a screen lock. When the screen is locked (e.g. by the pointer
* environment), the output buffer is emptied (its contents is discarded) and the
* screen is marked as garbage (so that MicroEMACS redraws it on the first occa-
* sion). If the screen is not locked then the buffer contents is flushed onto
* the screen.
*
_qlflush:
        tst.b   _insubmenu                      ; Is a MicroEMACS sub-menu open?
        bne     qlflush0                        ; If yes, then don't care lock.
        movem.l d3/a2,-(sp)                     ; Save the context.
        lea     dummy,a2                        ; Dummy external screen routine.
        move.l  _qlout,a0                       ; Console channel Id.
        moveq   #0,d3                           ; Don't wait too much (1/50 s).
        moveq   #sd.extop,d0                    ; Try to enable the cursor...
        trap    #3                              ;
        movem.l (sp)+,d3/a2                     ; Restore the context.
        tst.l   d0                              ; Error ?
        beq     qlflush0                        ; If no error, jump over...
        move.w  #1,_sgarbf                      ; Screen is garbage.
emptybuffer:
        move.w  #100,remains                    ; Empties the buffer.
        lea     qlbuffer,a0
        move.l  a0,pointer
        tst.l   d0                              ; Set CCR according to error.
        rts                                     ; Return.
qlflush0:
        moveq   #100,d2                         ; Size of buffer
        sub.w   remains,d2                      ; Find characters in buffer
        beq     return                          ; If none, skip
        move.w  #100,remains                    ; It's empty
        moveq   #io.sstrg,d0                    ; Send string
        lea     qlbuffer,a1                     ; Base of buffer
        move.l  a1,pointer                      ; Reset buffer pointer now
        bra     trap3

dummy:
        moveq   #0,d0
return:
        rts

*
* int qlmove(int row,int col)
* Change the cursor position
*
_qlmove:
        bsr     _qlflush                        ; Don't leave stuff in buffer
        bne     return
        moveq   #sd.pos,d0                      ; Reposition cursor
        move.l  8(sp),d1                        ; Column
        move.l  4(sp),d2                        ; Row
        bra     trap3

*
* int qleeol(void)
* Erase to end of line
* EMACS does not expect this to get reverse right, it only ever
* calls it when the logical and physical colours match
*
_qleeol:
        bsr     _qlflush                        ; Ensure no characters pending.
        bne     return
        moveq   #sd.setpa,d0                    ; Set paper colour
        move.b  _cbcolour,d1                    ; ...to background (a char)
        bsr     trap3
        moveq   #sd.clrrt,d0                    ; And arease to end of line.
        bra     trap3

*
* int _qleeop(void)
* Erase to end of page
* This is only ever called from within updgar() in display_c
* and is expected to set the entire screen to be the global
* background colour. Reverse status is irrelevant.
*
_qleeop:
        bsr     _qlflush                        ; Ensure no characters pending
        bne     ret
        moveq   #sd.setpa,d0
        move.l  _gbcolor,d1                     ; Get global background colour (an int)
        lea     _qlctrans,a0
        move.b  0(a0,d1.l),d1                   ; Get logical colour
        bsr     trap3
        moveq   #sd.clrrt,d0
        bsr     trap3
        moveq   #sd.clrbt,d0
        bra     trap3

*
* int _qlputc(int ch)
* Write a character into the buffer
* Handles ^H and ^M correctly.
*
_qlputc:
        move.b  7(sp),d1                        ; Character to write
        cmp.b   #8,d1                           ; ^H?
        beq     _qlputc_bs
        cmp.b   #13,d1                          ; ^M?
        beq     _qlputc_cr
        move.w  remains,d2                      ; Space left in buffer
        move.l  pointer,a0                      ; Address of space in buffer
        move.b  d1,(a0)+                        ; Insert character
        move.l  a0,pointer                      ; Reset pointer
        subq.w  #1,d2                           ; A little less space in buffer
        move.w  d2,remains                      ; Set space counter
        beq     _qlflush
ret:
        rts
_qlputc_bs:
        bsr     _qlflush                        ; Flush buffer
        bne     ret
        moveq   #sd.pcol,d0                     ; Go back 1 column
trap3:
        move.l  d3,-(sp)
        moveq   #forever,d3                     ; Wait until finished
        move.l  _qlout,a0                       ; Channel
        trap    #3                              ; Call IOSS
        move.l  (sp)+,d3
        tst.l   d0                              ; Test for errors.
        rts
_qlputc_cr:
        bsr     _qlflush                        ; Flush buffer
        bne     ret
        moveq   #sd.tab,d0                      ; Go to any column
        moveq   #0,d1                           ; Specifically, the leftmost column
        bra     trap3

*
* int _bktoshell(int f,int n)
* On UNIX systems, suspends EMACS and returns to a shell.
* On QDOS systems, with the pointer interface installed, this call allows
* you to pick any job (e.g. SuperBASIC); this is probably not very useful,
* but is included anyway.
* If called without arguments, then job 0,0 is assumed; otherwise, job with
* NUMBER n is used (e.g. <ESC>1^XD will pick job 1.)
*
_bktoshell:
        moveq   #mt.inf,d0                      ; Find system variables -> A0
        trap    #1
        moveq   #0,d1                           ; Default job ID
        move.l  4(sp),d0                        ; Get flag
        beq     pick_gotjob                     ; If flag = 0, use default job
        move.l  8(sp),d1                        ; Get job number
        cmp.w   sv_jbmax(a0),d1
        bhi     pick_bad                        ; Error if off top of table
        move.w  d1,d2
        lsl.w   #2,d2                           ; Offset in table
        move.l  sv_jbbas(a0),a0                 ; Base of table
        add.w   d2,a0                           ; Address of pointer
        move.l  (a0),d0                         ; Get address of job
        bmi     pick_bad                        ; If -ve, there is no job
        move.l  d0,a0
        swap    d1
        move.w  jb_tag(a0),d1                   ; Get job tag
        swap    d1                              ; ...giving job ID
pick_gotjob:
        moveq   #iop.pick,d0                    ; Pointer interface PICK call
        moveq   #0,d2                           ; No wake event
        bsr     trap3
        tst.l   d0
        bne     pick_bad                        ; Was there an error?
        moveq   #1,d0                           ; If no error return TRUE
        rts
pick_bad:
        moveq   #3,d0                           ; Error so return FAILED
        rts

*
* int _typahead()
* Check to see if there are any keystrokes waiting to be actioned
*
_typahead:
        move.l  d3,-(sp)
        moveq   #io.pend,d0                     ; Test for pending input
        moveq   #0,d3                           ; Don't bother waiting
        move.l  _qlout,a0
        trap    #3                              ; Do the test
        tst.l   d0                              ; 0 => input pending
        seq     d0                              ; ??????FF => input pending
        moveq   #1,d1
        and.l   d1,d0                           ; 1 => input pending
        move.l  (sp)+,d3
        rts

*
* Open the console
*
_qlopen:
        move.b  #1,_revexist
        jmp     _ttopen

*
* Close the console
*
_qlclose:
        pea.l   7
        jsr     _qlfcol
        clr.l   (sp)
        jsr     _qlbcol
        addq.l  #4,sp
        moveq   #sd.setpa,d0
        moveq   #0,d1
        bsr     trap3
        jmp     _ttclose

*
* (Re-)open keyboard
*
_qlkopen:
        moveq   #sd.cure,d0
        bsr     trap3
        move.b  _qlrevstate,d1
        bra     _qlrev0

*
* Close keyboard
*
_qlkclose:
ret2:
        rts

*
* Set foreground colour
*
_qlfcol:
        bsr     _qlflush
        bne     ret2
        move.l  4(sp),d1
        lea     _qlctrans,a0
        move.b  0(a0,d1.l),d1
        move.b  d1,_cfcolour
        moveq   #sd.setst,d0
        tst.b   _qlrevstate
        bne     trap3
        moveq   #sd.setin,d0
        bra     trap3

*
* Set background colour
*
_qlbcol:
        bsr     _qlflush
        bne     ret2
        move.l  4(sp),d1
        lea     _qlctrans,a0
        move.b  0(a0,d1.l),d1
        move.b  d1,_cbcolour
        moveq   #sd.setin,d0
        tst.b   _qlrevstate
        bne     trap3
        moveq   #sd.setst,d0
        bra     trap3

*
* Set reverse video state
*
_qlrev:
        bsr     _qlflush
        bne     ret2
        move.l  4(sp),d1
        move.b  d1,_qlrevstate
_qlrev0:
        moveq   #sd.setin,d0
        moveq   #sd.setst,d2
        tst.b   d1
        bne     _qlrev1
        exg     d0,d2
_qlrev1:
        move.b  _cbcolour,d1
        bsr     trap3
        move.b  _cfcolour,d1
        move.l  d2,d0
        bra     trap3

_qlunder:
        bsr     _qlflush
        bne     ret3
        move.l  4(sp),d1
        moveq   #sd.setul,d0
        bra     trap3

_qlbold:
        bsr     _qlflush
        bne     ret3
        move.l  4(sp),d1
        move.b  d1,_qlboldstate
        tst.b   _qlitalstate
        bne     ret3
        move.l  a2,-(sp)
        lea     _bold1,a1
        lea     _bold2,a2
        tst.b   d1
        bne     qlfount
qldeffnt:
        suba.l  a1,a1
        suba.l  a2,a2
qlfount:
        moveq   #sd.fount,d0
        bsr     trap3
qlfnt1:
        move.l  (sp)+,a2
ret3:
        rts

_qlitalic:
        bsr     _qlflush
        bne     ret3
        move.l  4(sp),d1
        move.b  d1,_qlitalstate
        move.l  a2,-(sp)
        lea     _italic1,a1
        lea     _italic2,a2
        tst.b   d1
        bne     qlfount
        tst.b   _qlboldstate
        beq     qldeffnt
        bra     qlfnt1

*
* Read mouse displacement
* int _qlmouse()
*
_qlmouse:
        moveq   #mt.inf,d0
        trap    #1
        movea.l sv_chbas(a0),a1
        move.l  _qlout,d1
        lsl.w   #2,d1
        movea.l 0(a1,d1.w),a1
        movea.l 4(a1),a1
        move.w  $2E(a1),d0
        rts

*
* Declare globals
*
        .globl  _qlflush,       _qlmove,        _qleeol,        _qleeop
        .globl  _qlputc,        _bktoshell,     _typahead,      _insubmenu
        .globl  _qlopen,        _qlclose,       _qlkopen,       _qlkclose
        .globl  _qlfcol,        _qlbcol,        _qlrev,         _qlmouse
        .globl  _qlunder,       _qlbold,        _qlitalic
